# %%
# Brat2DRT

import re
import glob
import os
import string

import platform

from pathlib import Path

# DRT
from nltk.sem import logic
from nltk.inference import TableauProver
from nltk.sem.drt import *
from text2story.brat2viz.brat import file_parser
# %%

dexpr = DrtExpression.fromstring

# ANN_DIR = 'brat_ann_files/'
#ANN_DIR = 'text_2_story_new/'
ANN_DIR = 'sample/'
#ANN_DIR = 'lusa_news/'
# ANN_DIR = 'lusa_news_old/'
#DRS_DIR = '../drs_files/'
DRS_DIR = 'drs_sample/'


def get_files(path=ANN_DIR):
    # path = r'.' # use your path
    path = r"{}".format(path)
    all_files = glob.glob(path + "/*.ann")
    print(all_files)
    return all_files


def read_file(filename):
    with open(filename, encoding="utf8") as f:
        content = f.readlines()

    content = [x.strip().replace('\t', ' ') for x in content]
    return content



# assign a variable for each event
def assign_variable(ann_dict):
    var = [chr(i) for i in range(ord('a'), ord('z') + 1)]
    var = var + [''.join([a, b]) for a in var for b in var]

    # events_var = []
    dexpr_list = []
    tmp = []
    i = 0
    for x in ann_dict:
        if x['ann_type'] == 'Event':
            if 'tag_ref' in x:
                dexpr_list.append(
                    ({'event_var': var[i], 'event_tag_id': x['tag_id'], 'tag_ref': x['tag_ref'],'event_str': x['value']},
                    dexpr('event(%s)' % var[i])))
            else:
                dexpr_list.append(
                        ({'event_var': var[i], 'event_tag_id': x['tag_id'],'event_str': x['value'],'tag_ref':'E'},
                    dexpr('event(%s)' % var[i])))
            i += 1
            x['var'] = var[i]

        elif x['ann_type'] == 'TextBound' and 'attribute' in x\
                and x['attribute'] == 'Event':
            # new version of annotation consider Event in a textbound tagid,
            # but to remain compatible with other annotations version the bellow condition
            # is still in the source code
            dexpr_list.append(
                ({'event_var': var[i], 'event_tag_id': x['tag_id'], \
                        'event_str': x['value'],'tag_ref':'E'},
                 dexpr('event(%s)' % var[i])))
            i += 1
            x['var'] = var[i]
        elif x['ann_type'] == 'TextBound' and 'attribute' in x\
                and x['attribute'] == 'EVENT':
            tmp.append((x['tag_id'], x['value']))

    for i, x in enumerate(dexpr_list):
        # in the new version of annotation we dont need a second list like tmp

        for y in tmp:
            if x[0]['tag_ref'] == y[0]:
                x[0]['event_str'] = y[1]
                dexpr_list[i] = x

    return dexpr_list, ann_dict


def attributes_events(dexpr_list, ann_dict):
    dr_set = set()  # discourse referents set

    for x in ann_dict:

        if x['ann_type'] == 'Attribute' and \
                x['tag_ref'][0] == 'E':
            for i, y in enumerate(dexpr_list):

                if y[0].get('event_tag_id') == x['tag_ref']:
                    if x['attr_type'] == 'Class':
                        y = y + \
                            (dexpr('event%s(%s,%s)' % (
                                x['attr_type'], y[0].get('event_var'), x['value'])),)
                    else:
                        y = y + \
                            (dexpr('%s(%s,%s)' % (x['attr_type'], y[0].get(
                                'event_var'), x['value'])),)

                    dexpr_list[i] = y
                    dr_set.add(x['value'])

    dexpr_list = format_tuple(dexpr_list)
    return dr_set, dexpr_list


def format_tuple(dexpr_list):
    for n, x in enumerate(dexpr_list):
        e_attr = list(x)
        k = x + (e_attr[1],)

        for i in e_attr[2:]:
            k = k + (k[-1] + i,)
        x = x + (k[-1],)
        dexpr_list[n] = x

    return dexpr_list


# get event ref variable
def get_ref(dexpr_list, var):
    ref = ''
    for i, y in enumerate(dexpr_list):
        if y[0].get('event_tag_id') == var:
            ref = (y[1], y[0].get('event_str'), y[0].get('event_var'))
            # drs1 = y[-1]
            break
    return ref


def get_value(ann_dict, var):

    for x in ann_dict:
        # apenas TextBound e Attribute
        if (x['ann_type'] == 'TextBound' or x['ann_type'] == 'Attribute') and (x['tag_id'] == var):
            return x.get('value')

def event_event_relation(ann_dict, dexpr_list):
    for x in ann_dict:

        if x['ann_type'] == 'Relation' and 'TLINK' in x['rel_type']:
            if 'before' not in x['rel_type'] and 'after' not in x['rel_type']:
                continue
            ref1 = ref2 = ''
            # drs1 = drs2 = ''

            ref2 = get_ref(dexpr_list, x['tag_ref2'])
            ref1 = get_ref(dexpr_list, x['tag_ref1'])

            for i, y in enumerate(dexpr_list):
                # print(i, ref1[2], type(ref1[2]),y[0].get('event_var'),x['rel_type'], type(x['rel_type']))
                if not ref1 or not ref2: continue
                if (y[0].get('event_var') == ref1[2]) and ('before' not in x['rel_type']):
                    n = len(y) - 2
                    y = y[: n] + (dexpr('occursAfter(%s,%s)' % (ref1[2], ref2[2])),) + y[n:]
                    dexpr_list[i] = y
                elif (y[0].get('event_var') == ref2[2]):
                    n = len(y) - 2
                    y = y[: n] + (dexpr('occursBefore(%s,%s)' % (ref2[2], ref1[2])),) + y[n:]
                    dexpr_list[i] = y

    return dexpr_list


def write_output(dexpr_list, dr_set, actors, actors_events, out_file):
    with open(out_file, 'w') as f:
        try:
            print('» EVENTS', file=f)

            for x in dexpr_list:

                if "event_str" not in x[0].keys():
                    continue

                var = dexpr('%s' % x[0].get('event_var'))
                event_str = ''.join([x.capitalize() for x in x[0].get('event_str').split()])
                #event_str = event_str.replace('-', '')
                event_str = "".join([c for c in event_str if c not in string.punctuation])
                event_str = event_str + '=' + x[0].get('event_var')
                var_ = dexpr('%s' % event_str)
                tmp_dr = []

                for y in dr_set:
                    tmp_dr = tmp_dr + [str(y) for x in list(x)
                    [1:-1] if re.search(y, str(x))]

                tmp_dr = [k for k in list(set(tmp_dr)) if type(k) is str]
                tmp_dr.append(var)


                print('# ' + x[0].get('event_tag_id') + ' (' +
                      x[0].get('event_str') + ') -> ' + x[0].get('event_var'), file=f)
                print('# FOL: ', DRS(tmp_dr, x[1:-1]).fol(), file=f)
                print('# DRS: ', DRS([var], x[1:-1]), '\n', file=f)
                print(DRS([var_], x[1:-1]).pretty_format())
        except Exception as e:
            print('Exception while writing EVENTS\n')
            print(e)

        try:
            print('» ACTORS', file=f)
            for x in actors:
                print('# ' + x[0] + ' -> ' + x[1], file=f)
        except Exception as e:
            print('Exception while writing ACTORS\n')
            print(e)

        try:
            print('\n» RELATIONS', file=f)
            for x in actors_events:
                print('# ' + x.get('ref1') + ' - ' + x.get('relation') + ' - ' + x.get('ref2'), file=f)
        except Exception as e:
            print('Exception while writing RELATIONS\n')
            print(e)


def get_actors(ann_dict):
    actors_dict = []
    for x in ann_dict:
        # new version of lusa news requires Participant as actors
        if x['ann_type'] == 'TextBound' and 'attribute' in x\
                and (x['attribute'] == 'ACTOR' or x['attribute'] == 'Participant'):
            d = {'tag_id': x['tag_id'], 'value': x['value']}
            actors_dict.append(d)
        elif x['ann_type'] == 'Participant':
            d = {'tag_id': x['tag_id'], 'value': x['value']}
            actors_dict.append(d)
    return actors_dict


def actors_relation(ann_dict):
    relations = list()
    for x in ann_dict:
        r_set = set()
        if x['ann_type'] == 'Relation':  # and x['rel_type'] == 'OBJ_REL_objIdentity':

            for y in ann_dict:
                if y['ann_type'] == 'Relation':  # and y['rel_type'] == 'OBJ_REL_objIdentity':
                    if (x['tag_ref1'] == y['tag_ref1']):
                        r_set.add(x['tag_ref1']);
                        r_set.add(y['tag_ref1'])
                    if (x['tag_ref1'] == y['tag_ref2']):
                        r_set.add(x['tag_ref2']);
                        r_set.add(y['tag_ref2'])
                    if (x['tag_ref2'] == y['tag_ref1']):
                        r_set.add(x['tag_ref1']);
                        r_set.add(y['tag_ref1'])
                    if (x['tag_ref2'] == y['tag_ref2']):
                        r_set.add(x['tag_ref2']);
                        r_set.add(y['tag_ref2'])
                relations.append(list(r_set))

    # tuple_list = list(set(tuple(i) for i in relations))
    # tuple_list = [x for x in tuple_list if len(x) > 1]
    # graph = nx.Graph(tuple_list)
    # result = list(nx.connected_components(graph))
    # print(result)

    tuple_list = list(set(tuple(i) for i in relations))
    tuple_list = [x for x in tuple_list if len(x) > 1]

    actors = get_actors(ann_dict)
    updated_actors = aux_actors(actors, tuple_list)
    updated_actors.sort()

    return updated_actors, tuple_list


# Seleciona somente atores que aparecem em uma relação
def aux_actors(actors, actors_rel):
    tuple_list = list()
    for y in actors_rel:
        for x in actors:
            if x.get('tag_id') in y:
                tuple_list.append((x.get('tag_id'), x.get('value')))
                # break
    return list(set(tuple_list))


def update_relations(actors, actors_sr, relations):
    new_relations = list()
    a = [i[0] for i in actors]
    t_list = set()
    # print(relations)
    for n, x in enumerate(relations):
        if not x['relation'] == 'objIdentity':
            r1 = {x.get('ref1')}
            for i in actors_sr:
                if r1.intersection(i):
                    import pdb
                    pdb.set_trace()
                    v = list(i.intersection(a))[0]
                    # print(v)
                    x['ref1'] = v
        new_relations.append(x)

    return new_relations


def update_actors(relations, actors, ann_dict):

    events_set = set()
    for elem in ann_dict:
        if 'attribute' in elem and\
                elem['attribute'] == "Event":
            events_set.add(elem["tag_id"])


    a = [i[0] for i in actors]
    r_list = list()
    for x in relations:
        if x['ref1'][0] == 'T' and x['ref1'] not in events_set:
            # print(x)
            r_list.append(x.get('ref1'))
        if x['ref2'][0] == 'T' and x['ref1'] not in events_set:
            # print(x)
            r_list.append(x.get('ref2'))

    new_list = list()
    for x in r_list:
        if x not in a:
            new_list.append(x)

    new_actors = list()
    for x in new_list:
        new_actors.append((x, get_value(ann_dict, x)))

    return actors + list(set(new_actors))


# actors_events relations
def TE_relations(ann_dict, dexpr_list):
    relations = list()
    for x in ann_dict:
        if x['ann_type'] == 'Relation' and \
                ('SEMROLE' in x['rel_type'] or 'SRLINK' in x['rel_type']):
            relations.append(
                {'ref1': x.get('tag_ref1'), 'ref2': x.get('tag_ref2'), 'relation': x.get('rel_type').split('_')[1]})

            event = x['tag_ref1'] if x['tag_ref1'][0] == 'E' else x['tag_ref2']
            var = x['tag_ref1'] if x['tag_ref1'][0] != 'E' else x['tag_ref2']

            event_ref = get_ref(dexpr_list, event)

            value = get_value(ann_dict, var)
            # value = ''.join([x.capitalize() for x in value.split(' ')])
            value = var

            try:
                for i, y in enumerate(dexpr_list):
                    if y[0].get('event_var') == event_ref[2]:
                        n = len(y) - 2
                        y = y[: n] + (dexpr('relationRole(%s,%s)' % (x.get('rel_type').split('_')[1], value)),) + y[n:]
                        dexpr_list[i] = y
            except:
                continue

    return relations, dexpr_list


def TT_relations(relations, ann_dict, dexpr_list):
    # relations = list()
    for x in ann_dict:
        # if x['ann_type'] == 'Relation' and 'OBJ_REL' in x['rel_type']:
        if x['ann_type'] == 'Relation' and ('OBJ_REL' in x['rel_type'] \
                or 'REF_REL' in x['rel_type'] or x['rel_type'].startswith('OLINK')):

            rel_type = x.get('rel_type').split('_')
            relations.append(
                {'ref1': x.get('tag_ref1'), 'ref2': x.get('tag_ref2'), 'relation':rel_type[-1]})

            # print(relations)
    return relations, dexpr_list


# %%
def process(file_name):

    print(file_name)
    filecontent = read_file(file_name)
    f_parser = file_parser(filecontent)
    dexpr_list, f_parser = assign_variable(f_parser)
    dr_set, dexpr_list = attributes_events(dexpr_list, f_parser)
    dexpr_list = event_event_relation(f_parser, dexpr_list)

    # actors = get_actors(f_parser)
    actors, actors_sr = actors_relation(f_parser)
    #print(actors)


    # platform.system() returns Windows, Linux or Darwin (for OSX)
    if platform.system() == 'Windows':
        output_file_name = file_name.split('\\')[-1].replace('.ann', '_drs.txt')
    else:
        output_file_name = file_name.split('/')[-1].replace('.ann', '_drs.txt')

    relations, dexpr_list = TE_relations(f_parser, dexpr_list)
    relations, dexpr_list = TT_relations(relations, f_parser, dexpr_list)

    # relations = update_relations(actors, actors_sr, relations)
    actors = update_actors(relations, actors, f_parser)

    write_output(dexpr_list, dr_set, actors, relations, output_file_name)


def main(ann_dir):
    files = get_files(ann_dir)
    # files = [files[3]]
    for f in files:
        print(f)
        filecontent = read_file(f)
        f_parser = file_parser(filecontent)
        dexpr_list, f_parser = assign_variable(f_parser)
        dr_set, dexpr_list = attributes_events(dexpr_list, f_parser)
        dexpr_list = event_event_relation(f_parser, dexpr_list)

        # actors = get_actors(f_parser)
        actors, actors_sr = actors_relation(f_parser)
        # print(actors)

        # Create output dir if does not exist
        Path(DRS_DIR).mkdir(parents=True, exist_ok=True)

        # platform.system() returns Windows, Linux or Darwin (for OSX)
        if platform.system() == 'Windows':
            output_file_name = f.split('\\')[1].replace('.ann', '_drs.txt')
        else:
            output_file_name = f.split('/')[1].replace('.ann', '_drs.txt')

        relations, dexpr_list = TE_relations(f_parser, dexpr_list)
        relations, dexpr_list = TT_relations(relations, f_parser, dexpr_list)

        # relations = update_relations(actors, actors_sr, relations)
        actors = update_actors(relations, actors, f_parser)

        output_file = os.path.join(DRS_DIR, output_file_name)
        write_output(dexpr_list, dr_set, actors, relations, output_file)

        # break


if __name__ == "__main__":
    main(ANN_DIR)

# %%



"""
	text2story.core.exceptions

	All defined exceptions raised by the package.
"""


class NoConnection(Exception):
	"""
	Raised if the user specified an invalid/unsupported language.
	"""

	def __init__(self):
		description = ('No internet Connection detected.')

		super().__init__(description)


class InvalidLanguage(Exception):
	"""
	Raised if the user specified an invalid/unsupported language.
	"""
	def __init__(self, lang):
		description = ('Invalid language :' + lang)
		
		super().__init__(description)


class InvalidNarrativeComponent(Exception):
	"""
	Raised if the user specified an invalid/unsupported narrative component.
	"""

	def __init__(self, element):
		description = ('Invalid narrative component :' + element)

		super().__init__(description)
class DuplicateNarrativeComponent(Exception):
	"""
	Raised if the user specified a duplicate narrative component.
	"""

	def __init__(self, element):
		description = ('Duplicate narrative component :' + element)

		super().__init__(description)


class InvalidTool(Exception):
	"""
	Raised if the user specified an invalid/unsupported tool.
	"""

	def __init__(self, tool):
		description = ('Invalid tool: ' + tool)
		
		super().__init__(description)



class InvalidLink(Exception):
	"""
	Raised if the user specified an invalid/unsupported tool.
	"""

	def __init__(self, link):
		description = ('Invalid link: ' + link)

		super().__init__(description)

class InvalidIDAnn(Exception):
	"""
	Raised if the user specified an invalid/unsupported tool.
	"""

	def __init__(self, entity_id):
		description = ('Invalid id: ' + entity_id)

		super().__init__(description)

class UninstalledModel(Exception):
    """
	Raised if the given model was not installed
    """

    def __init__(self, model_name, command):
        description = ('The follwing model was not installed: ' + model_name + \
					   '\n You should use the following command to install: ' + command)

        super().__init__(description)


class GraphBuilder:
    """
    # Usage
    narrative = Narrative(...)  # Create a narrative instance
    graph_builder = GraphBuilder(narrative)
    graph_builder.build_graph()

    """

    def __init__(self, narrative):
        self.narrative = narrative
        # Other initialization code if needed

    def build_graph(self):
        pass
        # Logic to read events, relations, and construct a graph using Narrative, EntityStructures, LinkStructures, etc.
        # Add nodes and edges to the graph based on the entities and relations




import os
import nltk
import argparse

from nltk.corpus import treebank, propbank
from nltk.chunk import tree2conlltags
from nltk.corpus.reader.propbank import PropbankChainTreePointer, PropbankSplitTreePointer

from text2story.core.narrative import Narrative
from text2story.readers.utils import convert_role_propbank_to_lirics, get_index

POS_NOUN = ["NN","NNP","NNS","NNP","NNPS","DT","JJ","JJS","JJR","PRP"]
TIME_EX = ["RB","RBR","RBS","IN","BY","CD"]

def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")


class ReadPropBank:

    def __init__(self):
        nltk.download("propbank")
        nltk.download("treebank")


    def convert_to_brat(self, lang, outputdir):

        self.build_txt_files(outputdir)
        self.build_ann_files(lang, outputdir)


    def get_index_pointer(self, pb_pointer):

        index_lst = []

        if isinstance(pb_pointer, PropbankChainTreePointer) or\
                isinstance(pb_pointer, PropbankSplitTreePointer):
            # a pointer can have discontinues elements, so it returns a list
            pieces_index = []

            # What to do with a chain?
            # iterate in the pieces and add to a list
            for piece in pb_pointer.pieces:
                pieces_index = pieces_index + self.get_index_pointer(piece)
            return pieces_index
        else:
            return [pb_pointer.wordnum]

    def count_offset(self, wordnum, sent_lst):

        count_offset = 0
        idx = 0
        while idx < wordnum:
            tok, pos, ne = sent_lst[idx]
            count_offset = count_offset + len(tok)
            idx = idx + 1

        return count_offset

    def get_time_args(self, argloc, ne_sent):
        index_lst = self.get_index_pointer(argloc)
        start = index_lst[0]

        if len(index_lst) > 1:
            end = index_lst[-1]
        else:
            end = len(ne_sent)

        count_offset = 0
        idx = 0
        # in this methodology, we have roughly the 
        # position of the token in text, in characters
        while idx < start:
            tok, pos, ne = ne_sent[idx]
            count_offset = count_offset + len(tok) 
            idx = idx + 1

        res = []
        while start < end:
            tok, pos, ne = ne_sent[start]
            if pos in TIME_EX:
                start = start + 1
            else:
                break


        return index_lst[0], start, count_offset

    def get_noun_args(self, argloc, ne_sent):
        index_lst = self.get_index_pointer(argloc)
        start = index_lst[0]

        if len(index_lst) > 1:
            end = index_lst[-1]
        else:
            end = len(ne_sent)

        count_offset = 0
        idx = 0
        # in this methodology, we have roughly the 
        # position of the token in text, in characters
        while idx < start:
            tok, pos, ne = ne_sent[idx]
            count_offset = count_offset + len(tok) 
            idx = idx + 1

        res = []
        while start < end:
            tok, pos, ne = ne_sent[start]
            if pos in POS_NOUN:
                start = start + 1
            else:
                break


        return index_lst[0], start, count_offset


    def _get_pointer_content(self, pointer, sent):


        if isinstance(pointer, PropbankChainTreePointer) or\
                isinstance(pointer, PropbankSplitTreePointer):
            # a pointer can have discontinues elements, so it returns a list
            pieces_txt = ""

            # What to do with a chain?
            # iterate in the pieces and add to a list
            for piece in pointer.pieces:
                pieces_txt = pieces_txt + self._get_pointer_content(piece, sent) + " "
            return pieces_txt.rstrip()
        else:
            tok_id = pointer.wordnum 
            tok_txt = sent[tok_id]
            return tok_txt


    def add_participant_narrative(self, narrative, pointer_content, offset, participant_set):

            offset_start = narrative.text.find(pointer_content, offset)
            offset_end = offset_start + len(pointer_content)

            if (offset_start, offset_end) not in participant_set:
                actor_id =  narrative._add_actor((offset_start, offset_end))
                participant_set.add((offset_start, offset_end))

                return actor_id, offset_end
            else:
                return None, -1

    def add_time_narrative(self, narrative, pointer_content, offset, time_set):

            offset_start = narrative.text.find(pointer_content, offset)
            offset_end = offset_start + len(pointer_content)

            if (offset_start, offset_end) not in time_set:
                time_id =  narrative._add_time((offset_start, offset_end))
                time_set.add((offset_start, offset_end))

                return time_id, offset_end
            else:
                return None, -1


    def add_arg_to_narrative(self, narrative, pointer_content, offset):

        if isinstance(pointer_content, tuple):
            tok_txt, ne_label = pointer_content

            offset_start = narrative.text.find(tok_txt, offset)
            offset_end = offset_start + len(tok_txt)
            actor_id = None
            if ne_label != 'O':
                actor_id = narrative._add_actor((offset_start, offset_end))

            return actor_id, offset_end
        else:
            # it is a list of tuples
            actor_lst = []
            current_offset = 0
            for content in pointer_content:
                actor_id, current_offset = self.add_arg_to_narrative(narrative, content, \
                        current_offset)
                if actor_id is not None:
                    actor_lst.append(actor_id)

            return actor_lst, current_offset

    def build_ann_files(self, lang, outputdir):

        narrative_instances = {}
        # the heuristic to map participants can add 
        # duplicate participants/events. Then this map helps to eliminate the 
        # repetead participants/events
        participants_instances = {} 
        events_instances = {} 
        times_instances = {} 

        for inst in propbank.instances():


            nameid = os.path.splitext(inst.fileid)[0]
            fname = os.path.join(outputdir, "%s.txt" % nameid)
            if not(os.path.exists(fname)):
                continue

            if nameid not in narrative_instances:
                with open(fname, "r") as fd:
                    txt = fd.read()
                    narrative_instances[nameid] = Narrative(lang, txt,"0000-00-00")
            narrative_inst = narrative_instances[nameid]
            # now, it is going to look for the arguments in the narrative text
            # to add events, actors, and srlinks 

            # id sentence of this propbank instance
            sent_id = inst.sentnum 
            # tokenized sentences from the txt file associated with this sentence
            sent_lst = treebank.sents(inst.fileid)  
            # pos tagged tokenized sentences from the txt file associated with this sentence
            tag_sent_lst = treebank.tagged_sents(inst.fileid)
            # entities of the tokenized sentences from the txt file associated with this sentence
            ne_sent = tree2conlltags(nltk.ne_chunk(tag_sent_lst[sent_id]))
            # the whole narrative text associated with the sentence sent_id
            narrative_txt = narrative_instances[nameid].text


            # add th ehead of the srl as event. Is that ok?
            event_current_offset = self.count_offset(inst.wordnum, ne_sent)
            event_txt = sent_lst[inst.sentnum][inst.wordnum]
            offset_event_start = narrative_txt.find(event_txt, event_current_offset)
            offset_event_end = offset_event_start + len(event_txt)

            if inst.fileid not in participants_instances:
                participants_instances[inst.fileid] = set()

            if inst.fileid not in events_instances:
                events_instances[inst.fileid] = set()

            if inst.fileid not in times_instances:
                times_instances[inst.fileid] = set()

            if (offset_event_start, offset_event_end) not in events_instances[inst.fileid]:
                event_id = narrative_instances[nameid]._add_event((offset_event_start, offset_event_end))
                events_instances[inst.fileid].add((offset_event_start, offset_event_end))
            else:
                event_id = None


            for argloc, argid in inst.arguments:

                lirics_role = convert_role_propbank_to_lirics(inst.roleset, argid)

                # start token index, end token index and
                # the position (char) in the file to start looking for the participants,
                # and events
                if argid == "ARGM-TMP": # it is a time expression
                    start, end, current_offset = self.get_time_args(argloc, ne_sent)
                    pointer_content = " ".join(sent_lst[sent_id][start:end])
                    time_id, time_offset = self.add_time_narrative(narrative_instances[nameid],\
                                    pointer_content,\
                                    0,\
                                    times_instances[inst.fileid])
                else:
                    start, end, current_offset = self.get_noun_args(argloc, ne_sent)
                    pointer_content = " ".join(sent_lst[sent_id][start:end])


                    if len(pointer_content) > 0:


                        actor_id, current_offset = self.add_participant_narrative(narrative_instances[nameid],\
                                pointer_content,\
                                current_offset,\
                                participants_instances[inst.fileid])

                        if actor_id is not None:
 
                            if lirics_role is not None and event_id is not None:
                                narrative_instances[nameid]._add_srlink(actor_id, event_id, "%s" % lirics_role)


        for nameid in narrative_instances:
            ann_fname = os.path.join(outputdir,"%s.ann" % nameid)

            with open(ann_fname, "w") as fd_ann:
                ann_txt = narrative_instances[nameid].ISO_annotation()
                fd_ann.write(ann_txt)



    def build_txt_files(self, outputdir):

        for fid in treebank.fileids():

            nameid = os.path.splitext(fid)[0]
            fname = os.path.join(outputdir, "%s.txt" % nameid)

            with open(fname, "w") as fd:

                raw_str = ""

                for sent in treebank.sents(fid):

                    for tok in sent:
                        if tok in ".,?!;":
                            raw_str = raw_str.rstrip()
                        raw_str = raw_str + tok + " "

                raw_str = raw_str.rstrip()

                fd.write(raw_str)

def main(lang, outputdir):

    reader = ReadPropBank()
    reader.convert_to_brat(lang, outputdir)

if __name__ == '__main__':
    my_parser = argparse.ArgumentParser(description='Convert PropBank corpus to brat format')

    my_parser.add_argument("language", action='store', type=str, help="The language to convert the propbank (pt or en).")
    my_parser.add_argument("datadir", action='store', type=dir_path, help="The directory where the brat files will be stored.")

    args = my_parser.parse_args()
    main(args.language, args.datadir)



import os
import json
working_dir = os.path.dirname(os.path.realpath(__file__))
pb_to_vb_map = json.load(open(os.path.join(working_dir,"pb-vn2.json"),"r"))
vb_to_lc_map = json.load(open(os.path.join(working_dir,"vn-lirics.json"),"r"))

straightforward_maps = {"ARGM-TMP":"time","ARGM-MNR":"manner"}

def get_index(lst, ele):
    # get the first occurrence index of ele in lst
    # if there is no element, return None
    for idx, e in enumerate(lst):
        if e == ele:
            return idx

    return None

def convert_role_propbank_to_lirics(roleset_prop, arg_label):

    # if the verb is not in the mapping, ignore

    if roleset_prop in pb_to_vb_map:

        # first it converts to verbnet, then it
        vn_mapping = pb_to_vb_map[roleset_prop].values()
        vb_label = None
        for v in vn_mapping:
            if arg_label.upper() in v:
               vb_label = v[arg_label.upper()]
               break

        if vb_label is not None and\
                vb_label in vb_to_lc_map.keys():
            return vb_to_lc_map[vb_label]

    if arg_label in straightforward_maps.keys():
        return straightforward_maps[arg_label]

    # if there is no mapping, just return None
    return None






from typing import List, Tuple
from text2story.readers.token_corpus import TokenCorpus

class Box:
    def __init__(self):
        self.variables = []
        self.box = []

    def __text2clauses(self, txt:str) -> List[Tuple[str,int,int]]:
        # TODO: write the function
        pass

    def __get_clause(self, tok:TokenCorpus, clause_lst:List[Tuple[str,int,int]]) -> Tuple[str,int,int]:
        # TODO: write the function
        # usar o atributo offset de tok para localizar qual oração
        # Busca binaria pela oracao a que o token pertence. Pesquisar tambem
        pass

    def __create_variable(self, entity:TokenCorpus)->Tuple[str,str,str]:
        pass

    def create_boxes(self, doc:List[TokenCorpus], doc_txt:str)-> List:
        clause_lst = self.__text2clauses(doc_txt)
        # TODO: ordenar esta lista de oracoes de acordo com o offset de inicio. Pesquisar
        # sobre ordenação de lista de tuplas
        box_lst = []
        current_box = Box()

        for tok in doc:

            # this means that tok has annotations
            if tok.attr != []: # [(ann_type, ann_map)] -> ('Event',{'Tense':'Past'...})
                variable = self.__create_variable(tok)
                clause = self.__get_clause(tok, clause_lst)
                # essa oração já foi analisada? Sim, então não preciso criar nova box. Não, crio nova box.
                # TODO 1): quando eu crio uma nova box. Atencao que pode ser recursiva!
                # TODO 2): adicionar a box a lista de box

                current_box.variables.append(variable)

        return box_lst







# Version of the text2story package
__version__ = "1.0.0"

from typing import List

from text2story.annotators import load
from text2story.annotators import get_available_tools
from text2story.annotators import get_participant_tools
from text2story.annotators import get_event_tools
from text2story.annotators import get_time_tools
from text2story.annotators import get_srlink_tools
from text2story.annotators import add_tool, get_tool_list

def get_tools(narrative_component:str):
    """
    get the tools for the given narrative component
    @param narrative_component:
    @return: a dicionary of tools and the languages it process
    """
    get_tool_list(narrative_component)

def add_annotator(tool_name:str, lang_lst:List[str], tool_types:List[str]):
    """
    It adds a tool to the gallery of annotators.

    @param tool: the name of the module where is your annotator
    @param lang_lst: a list of languages supported by your custom annotator,
                        for instance, ['en','fr']
    @param tool_types: a list of types that your annotator labels. The
                     supported types are ['participant','time', 'event','objectal_links',
                     'semantic_links']
    @return: None
    """

    add_tool(tool_name,lang_lst,tool_types)

def start(lang):
    load(lang)

def get_tools_name():
    return get_available_tools()

def participant_tool_names():
    return get_participant_tools()

def event_tool_names():
    return get_event_tools()

def time_tool_names():
    return get_time_tools()

def srlink_tool_names():
    return get_srlink_tools()


# Export to out of the package
from text2story.core.narrative import Narrative




"""
	text2story.core.narrative

	Narrative class
"""
import warnings

from text2story.core.annotator import Annotator
from text2story.core.entity_structures import *
from text2story.core.link_structures import *
from text2story.core.utils import pairwise, capfirst, join_tokens, find_first_non_space

import re

from text2story.core.utils import map_pos2head
class Narrative:
    """
    Representation of a narrative.

    Attributes
    ----------
    lang: str
            the language of the text; supported languages are portuguese ('pt') and english ('en')
    text: str
            the text itself
    publication_time : str
                    the publication time ('XXXX-XX-XX')
    participants: dict{str -> participant}
            the participants identified in the text.
            each key in the dict, of the form 'T' concatenated with some int, has an participant as a value.
    times: dict{str -> Time}
            the temporal expressions identified in the text.
            each key in the dict, of the form 'T' concatenated with some int, has an time as a value.
    obj_links: dict{str -> ObjectalLink}
            the corefs identified in the text
            each key in the dict, of the form 'R' concatenated with some int, has an coref as a value.

    Methods
    -------
    extract_participants(*tools)
            extracts all the participants in the text using the annotators defined in 'tools', updating self.participants
    extract_timexs(*tools)
            extracts all the timexs in the text using the annotators defined in 'tools', updating self.timexs
    extract_corefs(*tools)
            coreference resolution in the text using the tools 'tools', updating self.obj_rels
            typically, this call increases self.participants since news entities can be identified
    _get_participant_key(char_offset)
            returns the key of the participant with the corresponding character offset or None if such participant wasn't identified before
    _add_participant(char_offset)
            update self.participants by adding the new participant with character offset 'char_offset' and returns the key given to the new participant
    ISO_annotation(file_name)
            outputs ISO annotation in .ann format (txt)
    """

    def __init__(self, lang, text, publication_time):
        """
        Parameters
        ----------
        lang : str
                the language of the text
        text : str
                the text ifself
        publication_time : str
                the publication time ('XXXX-XX-XX')
        """

        self.lang = lang
        self.text = text
        self.publication_time = publication_time

        # Counter to generate a unique ID for every participant
        # TODO: Fix the counter, when repeting some extraction: The counter just keep going up.
        self._id = 1
        self._event_id = 1
        self._rel_id = 1

        self.participants = {}
        self.times = {}
        self.events = {}
        self.spatial_relations = {}

        self.links = {}

        self.sem_role_map = {"participant":"agent","event":"cause",\
                             "location":"location", "theme":"theme",\
                             "path":"path","manner":"manner","cause":"cause",\
                             "agent":"agent","goal":"goal"}
        self.sem_role_ignore = ["time"] # semantic links to ignore, there is no mapping
        #self.obj_links = {}
        #self.sem_links = {}

        #self.temp_links = {}
        #self.subordination_links = {}
        #self.qualitative_spatial_links = {}
        #self.movement_links = {}
        # self.measure_links = {}

    def get_relations_byargs(self, arg1, arg2):
        """
        Search for a relation entity whose arguments are arg1 and arg2
        @param arg1:
        @param arg2:
        @return: [link_structures], a set of relations (ids) between the two entities
        """
        relation_lst = []
        for relation_type in self.links:
            for relation in self.links[relation_type]:
                if relation.arg1 == arg1 and relation.arg2 == arg2:
                    relation_lst.append(relation)

                if relation.arg1 == arg2 and relation.arg2 == arg1:
                    relation_lst.append(relation)

        return relation_lst


    def get_entity_byname(self, name):
        return self.__getattribute__(name)

    @classmethod
    def fromTokenCorpus(cls, lang,  doc, raw_text=None):
        """
        build a narrative object from a list of TokenCorpus objects

        @return: a narrative object

        """

        cls.lang = lang
        cls.publication_time = None

        cls.participants = {}
        cls.times = {}
        cls.events = {}
        cls.spatial_relations = {}
        cls.links = {}

        # first collect only the annotations
        n = len(doc)
        i = 0
        map_id_tok = {}
        rel_set = set()
        doc_txt = ""
        while i < n:

            for idx, id_ann in enumerate(doc[i].id_ann):
                if id_ann in map_id_tok:
                    map_id_tok[id_ann] = (map_id_tok[id_ann][0] + [i],\
                                          map_id_tok[id_ann][1],\
                                          map_id_tok[id_ann][2])
                else:
                    ann_type, ann_map = doc[i].attr[idx]
                    map_id_tok[id_ann] = ([i], ann_type, ann_map)

            # collect relations
            for rel in doc[i].relations:
                if rel.rel_id not in rel_set:

                    rel_type = rel.rel_type.split("_")

                    main_type = rel_type[0]
                    subtype = rel_type[1]
                    if rel.argn == "arg2":
                        # this could produce an error, if the token has more than one id
                        # TODO: make a function that takes the id of arg1 for the given relation
                        # maybe I should chek in rel.toks[0].relations with the same id , and get the argument
                        arg1 = doc[i].id_ann[0]
                        arg2 = rel.toks[0].id_ann[0]
                    else:
                        # this could produce an error, if the token has more than one id
                        arg2 = doc[i].id_ann[0]
                        arg1 = rel.toks[0].id_ann[0]

                    linkObj = createLinkObject(arg1,arg2,main_type, subtype, rel.rel_id)
                    if main_type in cls.links:
                        cls.links[main_type].append(linkObj)
                    else:
                        cls.links[main_type] = [linkObj]

                    rel_set.add(rel.rel_id)

            doc_txt = doc_txt + doc[i].text + " "
            i += 1

        for id_ann in map_id_tok:
            # the tokens of this annotation
            idx_lst, ann_type, ann_map = map_id_tok[id_ann]
            text = [doc[i].text for i in idx_lst]
            text = join_tokens(text)

            span_start = doc[idx_lst[0]].offset
            span_end = doc[idx_lst[-1]].offset + len(doc[idx_lst[-1]].text)

            if ann_type == "Participant":
                # idx_lst[0].pos is the first pos tag option to label the lexical head of the participant
                lexical_head = ann_map.get("Lexical_Head", map_pos2head(doc[idx_lst[0]].pos))

                participant_type = ann_map.get("Participant_Type_Domain","")
                cls.participants[id_ann] = ParticipantEntity(text, (span_start, span_end), lexical_head, participant_type=participant_type)
            elif ann_type == "Event":
                event_class = ann_map.get("Class", "")
                event_polarity = ann_map.get("Polarity","Pos")
                event_type = ann_map.get("Event_Type","Occurrence")
                event_pos = ann_map.get("Pos","Verb")
                event_tense = ann_map.get("Tense","Pres")
                event_aspect = ann_map.get("Aspect","Perfective")

                cls.events[id_ann] = EventEntity(text, (span_start, span_end),event_class=event_class,\
                                                 polarity=event_polarity,event_type=event_type,pos=event_pos,\
                                                 tense=event_tense,aspect=event_aspect)
            elif ann_type == "Time":
                # annotations :: [(TimeStartOffset, TimeEndOffset, TimeType, TimeValue)]
                time_type = ann_map.get("Time_Type","Time")
                temporal_function = ann_map.get("TemporalFunction","Publication_Time")
                cls.times[id_ann] = TimeEntity(text,(span_start,span_end),
                                                                 time_type, temporal_function=temporal_function)
            elif ann_type == "Spatial_Relation":
                spatial_relation_type = ann_map.get("Type","Topological")
                topological = ann_map.get("Topological","Equal")
                pathDefining = ann_map.get("PathDefining","Starts")
                cls.spatial_relations[id_ann] = SpatialRelationEntity(text, (span_start, span_end),\
                                                                      type=spatial_relation_type,\
                                                                      topological=topological,
                                                                      pathDefining=pathDefining)

            else:
                warnings.warn("Annotation type not recognized: %s -> %s" % (id_ann, ann_type))



        if raw_text is None:
            narrativeSelf = cls(lang,doc_txt,"2023")
        else:
            narrativeSelf = cls(lang, raw_text, "2023")
        narrativeSelf.participants = cls.participants
        narrativeSelf.events = cls.events
        narrativeSelf.times = cls.times
        narrativeSelf.spatial_relations = cls.spatial_relations
        narrativeSelf.links = cls.links
        return narrativeSelf

    def extract_participants(self, *tools, url=None):
        """
        Parameters
        ----------
        tools : str, ...
                the tools to be used in the annotation

        Returns
        -------
                self.participants updated
        """

        participants = Annotator(tools).extract_participants(self.lang,
                                                 self.text,url)  # annotations :: [(EntityStartOffset, EntityEndOffset, EntityPOSTag, EntityType)]

        for participant in participants:
            self.participants['T' + str(self._id)] = ParticipantEntity(self.text[participant[0][0]:participant[0][1]], participant[0], participant[1],
                                                           participant[2])
            self._id += 1

        return self.participants

    def extract_times(self, *tools):
        """
        Parameters
        ----------
        tools : str, ...
                the tools to be used in the annotation

        Returns
        -------
                self.times updated
        """
        times = Annotator(tools).extract_times(
            self.lang, self.text, self.publication_time)  # annotations :: [(TimeStartOffset, TimeEndOffset, TimeType, TimeValue)]

        for time in times:
            self.times['T' + str(self._id)] = TimeEntity(self.text[time[0][0]:time[0][1]], time[0], time[1],
                                                         time[2])
            self._id += 1

        return self.times

    def extract_events(self, *tools):
        """
        Event extraction function to combine different tools of event extraction.
        Currently there is only one tool (AllenNLP) so it just uses that one.

        @param tools: Iterable with the tools to use

        @return: Returns a list of events extracted from the text in the form of EventEntity objects
        """
        events = Annotator(tools).extract_events(self.lang, self.text)
        if len(events) == 0:
            self.events = []
            return []

        for event in events.itertuples():
            self.events["T" + str(self._id)
                        ] = EventEntity(event.participant, event.char_span, event_class="Occurrence",\
                                        polarity="Pos", factuality="Factual",tense="Pres")
            self._id += 1

        return self.events

    def extract_objectal_links(self, *tools):
        """
        Parameters
        ----------
        tools : str, ...
                the tools to be used in the annotation

        Returns
        -------
                self.obj_rels updated
        """

        # annotations ::
        clusters = Annotator(tools).extract_objectal_links(
            self.lang, self.text)
        obj_links = []

        for cluster in clusters:
            for i in range(0, len(cluster) - 1):
                e1 = cluster[i]
                e2 = cluster[i + 1]

                # Get the participants
                arg1, arg2 = self._get_participant_key(e1), self._get_participant_key(e2)

                # If one of them wasn't identified in the participant extration, add it as a new participant
                if arg1 == None:
                    arg1 = self._add_participant(e1)
                if arg2 == None:
                    arg2 = self._add_participant(e2)

                # (Type (sameHead, partOf, ...), Arg1, Arg2)
                subtype = type='objIdentity'
                obj_links.append(ObjectalLink(arg1, arg2, subtype, "R" + str(self._rel_id)))
                self._rel_id += 1

        self.links["OLINK"] = obj_links
        return obj_links

    def extract_semantic_role_links(self, *tools):
        """
        Find semantic role links between extracted participants and events.
        Since the SRL model is different from the NER model, this function maps participants found by the SRL model into
        participants that were already extracted. If the participant was not yet extracted, it adds a new one.

        Links participants to events by text order. If we have participant1 -> EVENT1 -> participant2 in the text,
        then we make two semantic role links - EVENT1 -> ROLE -> participant1 | EVENT1 -> ROLE -> participant2

        @param tools: Iterable of tools to be used

        @return: A dict with the SRL entities by key -> R10: SemanticRoleLink<10>
        """
        srl_by_sentence = Annotator(
            tools).extract_semantic_role_links(self.lang, self.text)

        # FIND OUT IF ARGUMENT OF SRL HAS AN participant RETRIEVED BY THE NER COMPONENT
        # IF NOT, ADD A NEW participant CORRESPONDING TO THE ARGUMENT
        for sentence_df in srl_by_sentence:

            key_list = []
            for row in sentence_df.itertuples():
                event_key = None
                if row.sem_role_type == "EVENT":
                    event_key = self._get_event_key(row.char_span)
                    if event_key is not None:
                        key_list.append(event_key)
                    else:
                        event_key = self._get_event_key(
                            row.char_span, match_type="partial")
                        if event_key is not None:
                            key_list.append(event_key)
                        else:
                            event_key = self._add_event(row.char_span)
                            key_list.append(event_key)
                    continue
                participant_key = self._get_participant_key(
                    row.char_span, match_type="partial")

                if participant_key is not None:
                    key_list.append(participant_key)
                else:
                    if participant_key is None and event_key is None:
                        key_list.append("None")
                #else:
                #    participant_key = self._add_participant(
                #        row.char_span, lexical_head="Noun", participant_type="Other")
                #    key_list.append(participant_key)
            
            sentence_df["key"] = key_list

        # MAKE SEMANTIC ROLE LINK ENTITIES
        sem_links = []
        for sentence_df in srl_by_sentence:
            for row1, row2 in pairwise(sentence_df.itertuples()):
                sem1 = row1.sem_role_type
                sem2 = row2.sem_role_type

                if (sem1 == "EVENT") | (sem2 == "EVENT"):
                    if sem1 != "EVENT":
                        sem_role, participant, event = sem1, row1.key, row2.key
                    else:
                        sem_role, participant, event = sem2, row2.key, row1.key

                    if participant == "None" or event == "None":
                        continue
                    if sem_role not in self.sem_role_ignore:
                        sem_links.append(SemanticRoleLink(
                            participant, event, self.sem_role_map[sem_role.lower()], "R" + str(self._rel_id)))
                        self._rel_id += 1

        self.links["SRLINK"] =sem_links
        return sem_links

    def _add_srlink(self, participant, event, sem_role):
        self.links["SRLINKS"]["R" + str(self._rel_id)
                       ] = SemanticRoleLink(participant, event, self.sem_role_map[sem_role.lower()], "R" + str(self._rel_id))
        self._rel_id += 1

    def _get_participant_key(self, char_span, match_type="exact"):
        """
        Parameters
        ----------
        char_span : (int, int)
                the participant character offset of the participant to find the key
        match_type: str
                type of match for the character span.
                If exact, the span must be exactly the same to match an participant.
                If partial, the span must be partially contained in the participant span to match it.

        Returns
        -------
                the key of the participant with the corresponding character offset
                or None if it doesn't exist
        """
        if match_type == "exact":
            for key in self.participants.keys():
                if self.participants[key].character_span == char_span:
                    return str(key)
        elif match_type == "partial":
            for key in self.participants.keys():
                aSpan = self.participants[key].character_span
                if aSpan[0] <= char_span[0] <= aSpan[1]:
                    return str(key)
                elif aSpan[0] <= char_span[1] <= aSpan[1]:
                    return str(key)
        else:
            raise ValueError(
                f"Parameter match_type must be one of [exact, partial].\nInstead it was {match_type}")

        return None

    def _add_participant(self, char_span, lexical_head="Pronoun", participant_type="Other"):
        """
        Parameters
        ----------
        char_span : (int, int)
                the participant character offset
        lexical_head: str
                The lexical head of the participant: "Noun" or "Pronoun" -> Defaults to "Pronoun"
        participant_type: str
                The type of the participant as expressed by the NER models -> Defaults to "Other"

        Returns
        -------
                the key of the new added participant
        """
        key = 'T' + str(self._id)
        self.participants[key] = ParticipantEntity(self.text[char_span[0]:char_span[1]], char_span, lexical_head,
                                       participant_type)  # Hard-coded lexical head and type as 'Pronoun' and 'Other', resp., for now

        self._id += 1

        return key

    def _get_event_key(self, char_span, match_type="exact"):
        """
        Get the key of an event entity based on its character span on the full document text
        nota: Esta função está redundante com a _get_participant_key. Só muda a lista em que se procura

        @param char_span: The character span of the event in the text
        @return: The key of the event if it is found. None otherwise
        """

        if match_type == "exact":
            for key in self.events.keys():
                if self.events[key].character_span == char_span:
                    return str(key)
        elif match_type == "partial":
            for key in self.events.keys():
                aSpan = self.events[key].character_span
                if aSpan[0] <= char_span[0] <= aSpan[1]:
                    return str(key)
                elif aSpan[0] <= char_span[1] <= aSpan[1]:
                    return str(key)
        else:
            raise ValueError(
                f"Parameter match_type must be one of [exact, partial].\nInstead it was {match_type}")

        return None

    def _add_event(self, char_span):
        """
        Adds a new event to the narrative.

        @param char_span: tuple of characters (first_char, last_char) that delimit the event

        @return: The key of the new added event
        """
        key = 'T' + str(self._id)
        self.events[key] = EventEntity(
            self.text[char_span[0]:char_span[1]], char_span,  event_class="Occurrence",\
                                        polarity="Pos", factuality="Factual",tense="Pres")

        self._id += 1

        return key

    def _add_time(self, char_span):
        """
        Adds a new time expression to the narrative.

        @param char_span: tuple of characters (first_char, last_char) that delimit the time expression

        @return: The key of the new added time expression
        """
        key = 'T' + str(self._id)

        # TODO: change default values for value and times_type parameters
        self.times[key] = TimeEntity(
            self.text[char_span[0]:char_span[1]], char_span, "","Time")

        self._id += 1

        return key

    def _get_span_segments(self, el):
        """
        If a element of the narrative is composed of various segments (multiline annotations), then we have
        to collect each segment. Then it returns the character offsets of the segments and the segment text
        @param el:
        @return:
        """

        el_text = self.text[el.character_span[0]:el.character_span[1]]
        result = [_.start() for _ in re.finditer("\n", el_text)]

        span0 = el.character_span[0]
        span1 = el.character_span[0] + result[0]
        char_span_text = f"{span0} {span1};"
        span_text = f"{self.text[span0:span1]}"
        idx_result = 0

        while idx_result < len(result):

            span0 = str(el.character_span[0] + result[idx_result] + 1)
            if idx_result == len(result) - 1:
                span1 = str(el.character_span[1])
                char_span_text += span0 + " " + span1
            else:
                span1 = str(el.character_span[0] + result[idx_result + 1] - 1)
                char_span_text += span0 + " " + span1 + ";"

            # find the first non-space char to start the next text segment
            start_text_span = find_first_non_space(self.text, el.character_span[0] + result[idx_result])

            span_text += " " + f"{self.text[start_text_span:int(span1)]}"
            idx_result += 1

        return char_span_text, span_text

    def ISO_annotation(self):
        """
        Parameters
        ----------
                None

        Returns
        -------
                the ISO annotation in the .ann format
        """

        attribute_id = 1

        r = ""

        for participant_id in self.participants:
            participant = self.participants[participant_id]

            # brat does not deal well with multiline annotation, so
            # we need to split into several segments a multiline annotation
            if '\n' in participant.text:
                char_span_text, span_text = self._get_span_segments(participant)


                r += (participant_id + '\t' + 'Participant' + ' ' + char_span_text + '\t' + span_text + '\n')
            else:
                # T1 participant 0 22 O presidente de França
                span_text = self.text[participant.character_span[0]:participant.character_span[1]]
                r += (participant_id + '\t' + 'Participant' + ' ' + str(participant.character_span[0]) + ' ' + str(
                    participant.character_span[1]) + '\t' + span_text + '\n')

            # A1 Lexical_Head T1 Noun
            try:
                r += ('A' + str(attribute_id) + '\t' + 'Lexical_Head' +
                       ' ' + participant_id + ' ' + participant.lexical_head + '\n')
            except TypeError:
                print(attribute_id, type(attribute_id), participant_id,type(participant_id), \
                      participant.lexical_head, type(participant.lexical_head))

            attribute_id += 1

            # A2 Individuation T1 Individual
            r += ('A' + str(attribute_id) + '\t' + 'Individuation_Domain' +
                  ' ' + participant_id + ' ' + participant.individuation + '\n')
            attribute_id += 1

            # A3 participant_Type T1 Per
            r += ('A' + str(attribute_id) + '\t' + 'Participant_Type_Domain' +
                  ' ' + participant_id + ' ' + participant.type + '\n')
            attribute_id += 1

            # A4 Involvement T1 1
            r += ('A' + str(attribute_id) + '\t' + 'Involvement' +
                  ' ' + participant_id + ' ' + participant.involvement + '\n')
            attribute_id += 1

        for time_id in self.times:
            time = self.times[time_id]

            if '\n' in time.text:
                char_span_text, span_text = self._get_span_segments(time)

                r += (time_id + '\t' + 'Time' + ' ' + char_span_text + '\t' + span_text + '\n')
            else:
                # T26 TIME_X3 413 429 novembro de 2015
                span_text = self.text[time.character_span[0]:time.character_span[1]]
                r += (time_id + '\t' + 'Time' + ' ' + str(time.character_span[0]) + ' ' + str(
                    time.character_span[1]) + '\t' + span_text + '\n')

            # A55 Time_Type T26 Date
            # r += ('A' + str(attribute_id) + '\t' + 'Time_Type' + ' ' + time_id + ' ' + time.type + '\n')
            # attribute_id += 1

            # 6 AnnotatorNotes T26 value=2015-11-XX  ?????
            # r += ('A' + str(attribute_id) + '\t' + 'Value' + ' ' + time_id + ' ' + time.value + '\n')
            # attribute_id += 1

            # A107 FunctionInDocument T4 Publication_Time
            # r += ('A' + str(attribute_id) + '\t' + 'FunctionInDocument' + ' ' + time_id + ' ' + time.temporal_function + '\n')
            r += ('A' + str(attribute_id) + '\t' + 'TemporalFunction' +
                  ' ' + time_id + ' ' + time.temporal_function + '\n')
            attribute_id += 1

        for spatialRelation_id in self.spatial_relations:
            spatialRelation = self.spatial_relations[spatialRelation_id]

            r += (spatialRelation_id + '\t' + 'Spatial_Relation' + ' ' + str(spatialRelation.character_span[0]) + ' ' + str(
                spatialRelation.character_span[1]) + '\t' + spatialRelation.text + '\n')

            attr_dct = vars(spatialRelation)
            for att in attr_dct:
                if att != "character_span" and att != "text":
                    r +=(f"A{attribute_id}\t{capfirst(att)} {spatialRelation_id} {attr_dct[att]}\n")
                    attribute_id += 1

        for event_id in self.events:

            event = self.events[event_id]
            if len(event.text.strip()) == 0:
                continue

            if '\n' in event.text:
                char_span_text, span_text = self._get_span_segments(event)

                r += (event_id + '\t' + 'Event' + ' ' + char_span_text + '\t' + span_text + '\n')
            else:
                span_text = self.text[event.character_span[0]:event.character_span[1]]

                r += (event_id + '\t' + 'Event' + ' ' + str(event.character_span[0]) + ' ' + str(
                    event.character_span[1]) + '\t' + span_text + '\n')

            attr_dct = vars(event)
            for att in attr_dct:
                if att != "character_span" and att != "text":
                    r += (f"A{attribute_id}\t{capfirst(att)} {event_id} {attr_dct[att]}\n")
                    attribute_id += 1

            #r += (f"A{attribute_id}\tFactuality {event_id} {event.factuality}\n")
            #import pdb
            # pdb.set_trace()
            attribute_id += 1

        for link_type in self.links:
            for link in self.links[link_type]:
                r += str(link) + "\n"

        return r



from typing import List,Dict

def compose_ann_entity(ann_elements:List[str]) -> Dict[str, str]:
    """
    Given an annotation line of an entity, return a dictionary with its fields
    @return: the mapping of an annotation entity
    """
    if ';' in ann_elements[3]:
        ann = {'ann_type': 'TextBound', 'tag_id': ann_elements[0], 'attribute': ann_elements[1],
               's_pos': ann_elements[2], 'f_pos': ann_elements[4], 'value': ' '.join(ann_elements[5:]),
               'tag_ref': ann_elements[2]}
    else:
        ann = {'ann_type': 'TextBound', 'tag_id': ann_elements[0], 'attribute': ann_elements[1], \
               'tag': ann_elements[1].split(':')[0], 'value': ' '.join(ann_elements[4:]), \
               'tag_ref': ann_elements[2]}
    return ann

def file_parser(filecontent):
    ann_dict = []

    for x in filecontent:

        tmp = x.split()
        ann = {}
        if x[0] == 'E':
            ann = {'ann_type': 'TextBound', 'tag_id': tmp[0], 'tag': tmp[1].split(
                ':')[0], 'tag_ref': tmp[1].split(':')[1]}

        elif x[0] == 'A':
            ann = {'ann_type': 'Attribute',
                   'tag_id': tmp[0], 'attr_type': tmp[1], 'tag_ref': tmp[2], 'value': ' '.join(tmp[3:])}
        elif x[0] == 'T':
            ann = compose_ann_entity(tmp)
        elif x[0] == 'R':
            ann = {'ann_type': 'Relation', 'tag_id': tmp[0], 'rel_type': tmp[1], 'tag_ref1': tmp[2].split(':')[1],
                   'tag_ref2': tmp[3].split(':')[1]}
        elif x[0] == '#':
            ann = {'ann_type': 'Note',
                   'tag_id': tmp[0], 'tag_ref': tmp[1], 'note': ' '.join(tmp[2:])}

        if ann != {}:
            ann_dict.append(ann)

    return ann_dict
